import time
import threading

from zabbix_utils import Sender

from zbxepics.logging import logger
from zbxepics.casender.peekqueue import PriorityPeekQueue
from zbxepics.casender.item import MonitorItemFactory, IntervalItemFactory


class ZabbixSenderCA(object):
    """
    ZabbixSenderCA class, send metrics to Zabbix server.

    Attributes
    ----------
    _monitor_items : list of zbxepics.casender.item.MonitorItem
        list of monitor items
    _interval_item_q : zbxepics.casender.peekqueue.PriorityPeekQueue
        priority queue of interval items
    zbx_sender : pyzabbix.ZabbixSender
        Zabbix sender to send metrics to Zabbix server
    __is_stop : threading.Event
        whether server is stop or not
    __stop_request : bool
        to stop running server
    _is_running : bool
        whether server is running or not
    """

    def __init__(self, zabbix_server='127.0.0.1', zabbix_port=10051,
                 use_config=None, items=None):
        """
        Parameters
        ----------
        zabbix_server : str
            Zabbix server ip address (default is '127.0.0.1')
        zabbix_port : int
            Zabbix server port (default is 10051)
        use_config : str or bool
            Path to zabbix_agentd.conf file to load settings from.
            If value is 'True' then default config path will used:
            /etc/zabbix/zabbix_agentd.conf
        items: dict
            List of sender items.
            (Prerequisite keys: host pv interval, Optional: item_key func)
        """

        self._monitor_items = []
        self._interval_item_q = PriorityPeekQueue()
        self.zbx_sender = Sender(zabbix_server, zabbix_port, use_config)
        self.__is_stop = threading.Event()
        self.__stop_request = False
        self._is_running = False

        if items:
            for item in items:
                self.add_item(item)

    def add_item(self, item):
        """Add sender item to container

        Parameters
        ----------
        item : dict
            dict of item with following keys
            ('host', 'pv', 'interval', 'item_key', 'func')

        Returns
        -------
        item.MonitorItem or item.IntervalItem
            Added item
        """

        try:
            host = item['host']
            pvname = item['pv']
            interval = item['interval']
            item_key = item.get('item_key')

            if interval == 'monitor':
                sender_item = (MonitorItemFactory
                               .create_item(host, pvname, item_key))
                self._monitor_items.append(sender_item)
            else:
                func = item['func']
                sender_item = (IntervalItemFactory
                               .create_item(host, pvname, interval,
                                            func, item_key))
                self._interval_item_q.put((0, sender_item))
        except Exception:
            sender_item = None

        return sender_item

    def __get_interval_items(self):
        """Return a list of items to be executed

        Returns
        -------
        list of item.Intervalitem
            Return a list of items to be executed
        """

        if self._interval_item_q.empty():
            return []

        items = []
        now = int(time.time())
        while now >= self._interval_item_q.peek()[0]:
            _, item = self._interval_item_q.get()
            items.append(item)
            # Rescedule
            runtime = now + item.interval
            self._interval_item_q.put((runtime, item))

        return items

    def _create_metrics(self, items):
        """Return a list of metrics from item

        Parameters
        ----------
        items : list of item.MonitorItem or item.IntervalItem
            items to get metrics

        Returns
        -------
        list of pyzabbix.ZabbixMetric
            a list of metrics from item
        """

        metrics = []

        for item in items:
            try:
                zm = item.get_metrics()
                metrics.extend(zm)
            except Exception:
                pass

        return metrics

    def _send_metrics(self, items):
        """Send metrics to Zabbix server

        Parameters
        ----------
        items : list of item.MonitorItem or item.IntervalItem
        """

        metrics = self._create_metrics(items)
        if not metrics:
            return

        result = self.zbx_sender.send(metrics)
        logger.debug('%s: %s',
                     self.__class__.__name__,
                     result)

    def run(self):
        """Start ca sender server"""

        if (not self._monitor_items
                and self._interval_item_q.empty()):
            # Do not start if items is empty
            raise Exception('Sender process have no items.')

        self._is_running = True

        self.__is_stop.clear()
        try:
            while not self.__stop_request:
                items = []
                items.extend(self._monitor_items)
                items.extend(self.__get_interval_items())

                self._send_metrics(items)

                time.sleep(1)
        except Exception as err:
            logger.error(err)
        finally:
            logger.info('%s: %s',
                        self.__class__.__name__,
                        'Sender process stopped.')
            self.__stop_request = False
            self.__is_stop.set()

        self._is_running = False

    def stop(self):
        """Stop the run loop"""
        self.__stop_request = True
        self.__is_stop.wait()

    @property
    def is_running(self):
        """bool: Whether server is running or not"""
        return self._is_running



import time
import threading

import pytest

from . import zbxstreamserver
from zbxepics.casender import ZabbixSenderCA
from zbxepics.casender.item.monitoritem import MonitorItem
from zbxepics.casender.item.intervalitem import IntervalItem


def test_casender_init():
    sender = ZabbixSenderCA('testserver.com', 12345)

    zbx_sender = sender.zbx_sender
    host_port = str(zbx_sender.clusters[0])
    assert host_port == '[["testserver.com", 12345]]'


@pytest.mark.parametrize('interval,func,item_type', [
    ('monitor', None, MonitorItem),
    (10, 'last', IntervalItem),
])
def test_casender_add_item(interval, func, item_type):
    item = {'host': 'dummyServerHost', 'pv': 'ET_dummyHost:ai1'}
    item['interval'] = interval
    if func:
        item['func'] = func

    sender = ZabbixSenderCA('testserver.com', 12345)
    sender_item = sender.add_item(item)

    assert isinstance(sender_item, item_type)


def test_casender_add_item_err():
    item = {'host': 'dummyServerHost',
            'pv': 'ET_dummyHost:ai1'}
    sender = ZabbixSenderCA('testserver.com', 12345)
    sender_item = sender.add_item(item)

    assert sender_item is None


def test_casender_run_without_items():
    sender = ZabbixSenderCA('testserver.com', 12345)
    with pytest.raises(Exception):
        sender.run()


def test_casender_stop_request():
    item = {'host': 'dummyServerHost',
            'pv': 'ET_dummyHost:ai',
            'interval': 'monitor'}

    sender = ZabbixSenderCA('testserver.com', 12345)
    sender.add_item(item)

    th_sender = threading.Thread(target=sender.run)
    th_sender.daemon = True
    th_sender.start()

    time.sleep(1)
    assert sender.is_running is True
    sender.stop()
    assert sender.is_running is False

    th_sender.join()


def test_casender_sender_ca(softioc, caclient, zbx_stream):
    item = {'host': 'dummyServerHost',
            'pv': 'ET_dummyHost:long1',
            'interval': 'monitor'}

    zbx_host, zbx_port = zbx_stream
    sender = ZabbixSenderCA(zbx_host, zbx_port)
    sender_item = sender.add_item(item)

    th_sender = threading.Thread(target=sender.run)
    th_sender.daemon = True
    th_sender.start()

    pv = sender_item.pv
    pv.wait_for_connection(10)
    for i in range(5):
        pv.put(i, wait=True)
    time.sleep(1)

    sender.stop()
    th_sender.join()

    assert zbxstreamserver.metrics_received == 6



from queue import PriorityQueue


class PriorityPeekQueue(PriorityQueue):
    """PriorityQueue class with peek method"""

    def peek(self):
        """Return the first data.

        Returns
        -------
        obj
            the head of this queue.
            If the queue is empty, None is returned.
        """

        return None if self.empty() else self.queue[0]



import os
import logging
import logging.config


class Logger(object):
    """
    logging application log

    Attributes
    ----------
    logger : logging.Logger
        logger object
    """

    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    LEVELS = [NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL]

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        self.debug('Create logger(%s)', self.logger.name)

    def set_config(self, config_file):
        """Set configuration for logging from config file

        Parameters
        ----------
        config_file : str or bool
            Path to config file to load settings from.
            If value is `True` then default config path will used.
        """

        # If value is `True` then default config path will used
        if config_file and isinstance(config_file, bool):
            dir_path = os.path.dirname(__file__)
            config_file = os.path.join(dir_path, 'logging.conf')
        # Set configuration for logging
        logging.config.fileConfig(config_file,
                                  disable_existing_loggers=False)

        self.debug('Changed configuration for logger(%s)', self.logger.name)

    def set_level(self, level):
        """Set the logging level of this logger

        Parameters
        ----------
        level : int
            logging level
        """

        if level in self.LEVELS:
            self.logger.setLevel(level)

    def debug(self, msg, *args, **kw):
        """Log 'msg % args' with severity 'DEBUG'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.log(self.DEBUG, msg, *args, **kw)

    def info(self, msg, *args, **kw):
        """Log 'msg % args' with severity 'INFO'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.log(self.INFO, msg, *args, **kw)

    def warning(self, msg, *args, **kw):
        """Log 'msg % args' with severity 'WARNING'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.log(self.WARNING, msg, *args, **kw)

    def error(self, msg, *args, **kw):
        """Log 'msg % args' with severity 'ERROR'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.log(self.ERROR, msg, *args, **kw)

    def critical(self, msg, *args, **kw):
        """Log 'msg % args' with severity 'CRITICAL'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.log(self.CRITICAL, msg, *args, **kw)

    def log(self, level, msg, *args, **kw):
        """Log 'msg % args' with the integer severity 'level'.

        Notes
        -----
            To pass exception information, use the keyword argument
            exc_info with a true value.
        """

        self.logger.log(level, msg, *args, **kw)


logger = Logger()



import threading

from epics import PV
from zabbix_utils import ItemValue


class IntervalItem(object):
    """
    a class for interval item

    Attributes
    ----------
    host : str
        host name of item
    pvname : str
        pv name to monitor
    pv : epics.PV
        pv object to be monitored
    item_key : str
        item key of item
    interval : float
        interval in senconds to send
    value : obj
        value to be got
    __last_value : obj
        last got value
    _lock : threading.Lock()
        lock for value
    """

    DEFAULT_INTERVAL = 30.0

    def __init__(self, host, pvname, interval=30.0, item_key=None):
        """
        Parameters
        ----------
        host : str
            host name of item
        pvname : str
            pv name to monitor
        interval : float
            interval in senconds to send (default is 30.0)
        item_key : str
            item key of item (default is None)
        """

        self.host = str(host)
        pvname_ = str(pvname)
        self.pv = PV(pvname_,
                     connection_callback=self._on_connection_change,
                     callback=self._on_value_change)
        self.item_key = pvname_ if item_key is None else item_key

        self.interval = interval
        if self.interval is None or self.interval < 1.0:
            self.interval = self.DEFAULT_INTERVAL

        self._value = None
        self.__last_value = None

        self._lock = threading.Lock()

        self._setup()

    def _setup(self):
        """May be overridden by a subclass"""
        pass

    def _on_connection_change(self, pvname=None, conn=None, **kws):
        """Callback to be called every connection state change

        Parameters
        ----------
        pvname : str
            the name of the pv
        conn : bool
            specifying whether the PV is now connected

        Notes
        -----
            May be overridden by a subclass.
        """

        if not conn:
            self._value = None
            self._setup()

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin

        Notes
        -----
            May be overridden by a subclass.
        """

        pass

    def __get_value(self):
        """Get current value

        Returns
        -------
        obj
            current value if value is updated else last value
        """

        if not self.pv.connected:
            return None

        with self._lock:
            if self._value is not None:
                self.__last_value = self._value
                self._value = None
                self._setup()

        return self.__last_value

    def get_metrics(self):
        """Get Zabbix Metircs

        Returns
        -------
        list of pyzabbix.ZabbixMetric
            list of zabbix metric
        """

        value = self.__get_value()
        if value is None:
            return []

        zm = ItemValue(self.host, self.item_key, value)
        return [zm]

    def __lt__(self, other):
        return True


class IntervalItemLast(IntervalItem):
    """a class for last interval item"""

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin
        """

        with self._lock:
            self._value = value


class IntervalItemMin(IntervalItem):
    """a class for min interval item"""

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin
        """

        with self._lock:
            self._value = (value if self._value is None
                           else min(self._value, value))


class IntervalItemMax(IntervalItem):
    """a class for max interval item"""

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin
        """

        with self._lock:
            self._value = (value if self._value is None
                           else max(self._value, value))


class IntervalItemAvg(IntervalItem):
    """
    a class for average interval item

    Attributes
    ----------
    sum : int or float
        sum of values
    count : int
        counts of value change
    """

    def _setup(self):
        """Set up to initialize"""
        self.__sum = 0
        self.__count = 0

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin
        """

        with self._lock:
            self.__sum += value
            self.__count += 1
            self._value = float(self.__sum) / self.__count



from zbxepics.casender.peekqueue import PriorityPeekQueue


def test_peekqueue_init():
    q = PriorityPeekQueue()
    assert q is not None


def test_peek_queue_refer():
    """
    Test for refer to elements with the highest priority
     without removing them.
    """
    q = PriorityPeekQueue()
    for val in range(5):
        q.put(val)

    peek_value = q.peek()
    assert peek_value == 0
    assert q.qsize() == 5


def test_peek_queue_err():
    q = PriorityPeekQueue()
    peek_value = q.peek()

    assert peek_value is None



import time

import pytest
from epics import PV

from zbxepics.casender.item import MonitorItemFactory, IntervalItemFactory
from zbxepics.casender.item.monitoritem import MonitorItem
from zbxepics.casender.item.intervalitem import IntervalItem


def test_casender_item_init_monitor():
    item = (MonitorItemFactory
            .create_item('host1', 'ET_dummyHost:ai1'))

    assert isinstance(item, MonitorItem)
    assert item.host == 'host1'
    assert isinstance(item.pv, PV)
    assert item.pv.pvname == 'ET_dummyHost:ai1'


def test_casender_item_init_interval():
    item = (IntervalItemFactory
            .create_item('host1', 'ET_dummyHost:ai1',
                                  5, 'last'))

    assert isinstance(item, IntervalItem)
    assert item.host == 'host1'
    assert isinstance(item.pv, PV)
    assert item.pv.pvname == 'ET_dummyHost:ai1'
    assert item.interval == 5


def test_casender_item_init_interval_default():
    item = (IntervalItemFactory
            .create_item('host1', 'ET_dummyHost:ai1'))

    default_interval = IntervalItem.DEFAULT_INTERVAL
    assert item.interval == default_interval

    functions = IntervalItemFactory.list_of_functions()
    default_function = IntervalItemFactory.DEFAULT_FUNCTION
    assert type(item) == functions[default_function]


def test_casender_item_monitor_item_metrics(softioc, caclient):
    item = (MonitorItemFactory
            .create_item('host1', 'ET_dummyHost:long1'))

    pv = item.pv
    pv.wait_for_connection(10)
    test_vals = [v for v in range(5)]
    for val in test_vals:
        pv.put(val, wait=True)
    time.sleep(.05)

    metrics = item.get_metrics()
    assert len(metrics) == 6

    for (zm, tval) in zip(metrics[1:0], test_vals):
        assert zm.host == item.host
        assert zm.key == item.item_key
        assert zm.value == str(tval)


@pytest.mark.parametrize('func,values,result', [
    ('last', [0, 1, 2, 3, 4], '4'),
    ('min', [0, 1, 2, 3, 4], '0'),
    ('max', [0, 1, 2, 3, 4], '4'),
    ('avg', [0, 1, 2, 3, 4], '2.0'),
])
def test_casender_item_interval_item(softioc, caclient, func, values, result):
    item = (IntervalItemFactory
            .create_item('host1', 'ET_dummyHost:long1', func=func))

    pv = item.pv
    pv.wait_for_connection(10)
    for val in values:
        pv.put(val, wait=True)
    time.sleep(.05)

    metrics = item.get_metrics()
    assert len(metrics) == 1
    assert metrics[0].host == item.host
    assert metrics[0].key == item.item_key
    assert metrics[0].value == result

    pv.put(0, wait=True)
    pv.disconnect()
    metrics = item.get_metrics()
    assert metrics == []



from . import monitoritem
from . import intervalitem


class MonitorItemFactory:
    """Factory for creating monitor item"""

    @classmethod
    def create_item(cls, host, pvname, item_key=None):
        """Create monitor item

        Returns
        -------
        item.MonitorItem
            created item
        """

        return monitoritem.MonitorItem(host, pvname, item_key)


class IntervalItemFactory:
    """Factory for creating interval item"""

    DEFAULT_FUNCTION = 'last'

    _functions = {'last': intervalitem.IntervalItemLast,
                  'min': intervalitem.IntervalItemMin,
                  'max': intervalitem.IntervalItemMax,
                  'avg': intervalitem.IntervalItemAvg}

    @classmethod
    def create_item(cls, host, pvname, interval=30.0,
                    func='last', item_key=None):
        """Create interval item

        Parameters
        ----------
        host : str
            host name of item
        pvname : str
            pv name to monitor
        interval : float
            interval in senconds to send  (default is 30)
        func : str
            function to apply item buffer (default is 'last')
        item_key : str
            item key of item (default is None)

        Returns
        -------
        item.IntervalItem
            created item
        """

        if (func not in cls._functions):
            # Set default function
            func = cls.DEFAULT_FUNCTION

        item_class = cls._functions[func]
        return item_class(host, pvname, interval, item_key)

    @classmethod
    def list_of_functions(cls):
        """get a list of functions

        Returns
        -------
        list
            list of functions
        """

        return cls._functions



from queue import Queue

from epics import PV
from zabbix_utils import ItemValue


class MonitorItem(object):
    """
    a class for monitor item

    Attributes
    ----------
    host : str
        host name of item
    pvname : str
        pv name to monitor
    pv : epics.PV
        pv object to be monitored
    item_key : str
        item key of item
    __metric_q : queue.Queue
        queue of ItemValue
    """

    def __init__(self, host, pvname, item_key=None):
        """
        Parameters
        ----------
        host : str
            host name of item
        pv : epics.PV
            pv to monitor
        item_key : str
            item key of item
        """

        self.host = str(host)
        pvname_ = str(pvname)
        self.pv = PV(pvname_, callback=self._on_value_change)
        self.item_key = pvname_ if item_key is None else item_key

        self.__metrics_q = Queue()

    def _on_value_change(self, value=None, timestamp=None, **kw):
        """Callback to be called every monitor update

        Parameters
        ----------
        value : obj
            updated value
        timestamp : float
            timestamp of pv relative to POSIX time origin
        """

        zm = ItemValue(self.host, self.item_key,
                          value, int(timestamp))
        self.__metrics_q.put(zm)

    def get_metrics(self):
        """Get metrics

        Returns
        -------
        list of pyzabbix.ItemValue
            All metrics in queue
        """

        metrics = []
        while not self.__metrics_q.empty():
            metrics.append(self.__metrics_q.get())
        return metrics



from setuptools import find_packages, setup

with open('README.md') as f:
    readme = f.read()

setup(
    name='zabbix-epics-py',
    version='1.1.0',
    url='https://github.com/sasaki77/zabbix-epics-py',
    maintainer='Shinya Sasaki',
    description='Zabbix-EPICS for Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=('tests', 'scripts')),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'future',
        'zabbix_utils',
        'pyepics',
        ],
    extras_require={
        'develop': [
            'pytest',
            'pytest-cov',
            'pycodestyle'
            ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
